#include <unistd.h>
#include "mpi.h"
#include <cstdlib>
#include <cmath>
#include <cstring>
#include <iostream>
#include <fstream>
#include "ti/fftlib/src/common/fft_common.h"

#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>
#include "ocl_util.h"
using namespace cl;
using namespace std;

/*=============================================================================
* sizeof(fft_plan_t) will not suffice, since a pointer on the host and a pointer
*    on the device may be different sizes
*============================================================================*/
#define FFT_PLAN_T_SIZE (sizeof(fft_plan_t))
#define FFTSZ (64*1024)

#define XSTR(x) #x
#define STR(x) XSTR(x)

/* Output format string to store real and imaginary part of the output */
char csv_output[FFTSZ][50];
/* ======================================================================== */
/*  MAIN -- Top level driver for the test.                                  */
/* ======================================================================== */
int main  (int argc, char *argv[])
{
  int i, j;
  int rad = 4;
  int rank, num_processes;
  char processor_name[320];
  int name_len;
  int errs = 0;
  int size, rc, *buf_in, *buf_out, count;
  MPI_File fh_in, fh_out;
  MPI_Comm comm;
  MPI_Status status;
  uint32_t nodeBits,flags;
   
  /* Read parameters from command line */
   if(argc < 3)
  {
    printf("Usage: %s <input_file> <output_file>\n",argv[0]);
    exit(-1);
  }
 /* Initiate the MPI process over multiple nodes. In this demo there is one MPI process per chip */
  MPI_Init( &argc, &argv );
  comm = MPI_COMM_WORLD;
  MPI_Comm_size( comm, &num_processes );
  MPI_Comm_rank( comm, &rank );
  int total_size,chunks_per_chip,num_chips,num_iterations;

  /* Get the name of the processor */
  MPI_Get_processor_name(processor_name, &name_len);
  gethostname(processor_name, 320);

  printf ("\n %s:.... FFT execution begins.....", processor_name);
  fflush(stdout);
 
  /* Test the input file */
  MPI_File_open( comm, argv[1], MPI_MODE_RDONLY, MPI_INFO_NULL, &fh_in );
  MPI_File_open( comm, argv[2], MPI_MODE_WRONLY | MPI_MODE_CREATE, MPI_INFO_NULL, &fh_out );

   try
   {
     Context context(CL_DEVICE_TYPE_ACCELERATOR);
     std::vector<Device> devices = context.getInfo<CL_CONTEXT_DEVICES>();

     int d = 0;
     int bufsize = sizeof(float) * (FFTSZ);

     /* Kernels only read from and write to these DDR buffers */
     Buffer bufX   (context, CL_MEM_READ_WRITE, bufsize);
     Buffer bufY   (context, CL_MEM_READ_WRITE, bufsize*2);
     Buffer bufW   (context, CL_MEM_READ_WRITE, bufsize);
     Buffer bufPlan(context, CL_MEM_READ_WRITE, FFT_PLAN_T_SIZE);

     /*------------------------------------------------------------------------
     * Read the offline compiled kernel module
     *-----------------------------------------------------------------------*/
     char *bin;
     int bin_length = ocl_read_binary("../c66x/kernel.out", bin);

     Program::Binaries   binary(1, std::make_pair(bin, bin_length));
     Program             program = Program(context, devices, binary);
     program.build(devices);

     delete [] bin;

     /*------------------------------------------------------------------------
     * Create a Kernel from program, function name: ocl_FFT_plan_1d_sp_r2c
     *-----------------------------------------------------------------------*/
     Kernel FFT_plan   (program, "ocl_FFT_plan_1d_sp_r2c");

     /*------------------------------------------------------------------------
     * Set Kernel arguments for indices: 0-6
     * When setArg takes a Buffer, the corresponding function argument points
     * to the memory in the Buffer
     *-----------------------------------------------------------------------*/
     FFT_plan.setArg(0, FFTSZ);
     FFT_plan.setArg(1, bufX);
     FFT_plan.setArg(2, bufY);
     FFT_plan.setArg(3, bufW);
     FFT_plan.setArg(4, (int)0);
     FFT_plan.setArg(5, (int)FFT_DIRECT);
     FFT_plan.setArg(6, bufPlan);

     Kernel FFT_execute(program, "ocl_FFT_execute");
     FFT_execute.setArg(0, bufPlan);

     /*------------------------------------------------------------------------
     * Create a command queue, note that this is an IN ORDER queue
     *-----------------------------------------------------------------------*/
     CommandQueue Q(context, devices[d], CL_QUEUE_PROFILING_ENABLE);

     Event e;

     float *in;
     float *out;

     bool continue_computing=true;
     int iteration=0;
     /* The input file is read in terms of 64K sized blocks of data, and dispatched to the DSP for
        FFT processing. If there are 'n' DSPs avilable, n*64K data is read from the input file and 
       64K blocks are dispatched to the n DSPs for concurrent processing. The dispatching mechanism
       happens via MPI processes, one of which run on each of the n DSPs. Those MPI processes, with
       the help of OpenCL executes the FFT algorithm on its DSP cores and store the result in a shared
       file.
       The input file is shared across the MPI processes. As mentioned earlier, a batch of n*64K bytes
       are read from the input file. The file read pointer for an MPI process is moved to the beginning
       of the 64K data chunk, which it will read and perform FFT on. (The 'rank' gives the index of the
       process among the 'n' processes (num_processes) which could be used to advance the file pointer)
       Once  the execution is done, the output is written to the output file by the processes, in its
       respective positions within the file.
       This iteration is continued until the end of file is reached.
     */  


     while(1) {
       /* Map bufX into host's virutal address space for WRITE, CL_TURE blocks
          the call until MapBuffer finishes */
       in = (float*) Q.enqueueMapBuffer(bufX, CL_TRUE, CL_MAP_WRITE, 0, bufsize);
       
       /* Set the file pointers to the position within the num_processes*64K block */
       MPI_File_seek(fh_in, (iteration* num_processes + rank) * sizeof(float) * FFTSZ, MPI_SEEK_SET);

       memset( &status, 0xff, sizeof(MPI_Status) );
       MPI_File_read( fh_in, in, FFTSZ, MPI_FLOAT, &status );
       MPI_Get_count( &status, MPI_INT, &count );

       if(count<FFTSZ) {
         printf("\n %s: End of file reached, after %d iterations",processor_name,iteration);
         break;
       }
   
       /* Unmap bufX (write data back), free the virtual address space */
       Q.enqueueUnmapMemObject(bufX, in);

       /* Dispatch FFT_plan to device. Since Q is an in-order queue,
          only one task runs on device at a time.  However, FFT_plan
          task itself might contain OpenMP code, FFT_plan could still
          run parallel on multiple cores on device.
          Track it with an event. */
       Q.enqueueTask       (FFT_plan, 0, &e);
       /* Wait for FFT_plan to finish */ 
       e.wait();
       /* Dispatch FFT_execute to device. See notes above, FFT_execute
          might contain OpenMP code and run parallel on cores. */
       Q.enqueueTask       (FFT_execute, 0, &e);
       /* Wait for FFT_execute to finish */ 
       e.wait();
       
 
       cl_ulong t_strt, t_end;
       e.getProfilingInfo(CL_PROFILING_COMMAND_START,  &t_strt);
       e.getProfilingInfo(CL_PROFILING_COMMAND_END,    &t_end);
       cout << "\n"<< processor_name <<":" << " FFTW execute: "<<"iteration: "<<iteration<<" executed in " << ((t_end - t_strt)/1000) << " us" << endl;
 
       /* Map bufY into host's virtual address space for READ, CL_TRUE blocks
          call until MapBuffer finishes */
       out = (float*) Q.enqueueMapBuffer(bufY, CL_TRUE, CL_MAP_READ, 0, bufsize);
       /* The output (out[]) is a FFTSZ sized floating point pair or real & imaginary values (x,y).
          THis output (out[]) is formated in to strings containing the comma seperated values of
          the real and imaginary part and these output strings are stored in to the outputfile using MPI */ 
        for(i=0;i<FFTSZ;i++) {
          sprintf(csv_output[i],"%e,%e\n",out[2*i],out[2*i + 1]);
        } 
       /* Seting the file pointer to the position within the output file */
       MPI_File_seek(fh_out, (iteration* num_processes + rank) * sizeof(csv_output), MPI_SEEK_SET );
       /* Store the comma seperated formatted fft output in to the output file */
       MPI_File_write( fh_out, csv_output,sizeof(csv_output), MPI_CHAR, &status );

       /* Unmap bufY, free virtual address space */
       Q.enqueueUnmapMemObject(bufY, out);

       iteration++;
    } 
     printf("\n %s .....FFT execution completed....\n", processor_name);
  }
  catch (Error err)
   { cerr << "ERROR: " << err.what() << "(" << err.err() << ")" << endl; }
 
   /* Making sure all processes are completed up to this point */
    MPI_Barrier( comm );

    MPI_File_close( &fh_in );
    MPI_File_close( &fh_out );
    MPI_Finalize();
 
}
